# from linora.param_search import GEClassifier
# from linora.param_search import GERegressor
from linora.param_search import LGBMClassifier
from linora.param_search import LGBMRegressor
from linora.param_search import XGBClassifier
from linora.param_search import XGBRegressor
# from linora.param_search import XGBRanker

from linora.param_search._HyperParameters import *
from linora.param_search._Search import *

from linora.param_search._config import __xgboost_version__, __lightgbm_version__




import numpy as np

__all__ = ['normal_loss', 'mean_absolute_error', 'mean_squared_error',
           'mean_absolute_percentage_error', 'hinge', 'explained_variance_score',
           'median_absolute_error', 'r2_score', 'report_regression',
           'mean_relative_error', 'poisson', 'log_cosh_error'
          ]


def normal_loss(y_true, y_pred, k, log=False, root=False, sample_weight=None):
    """Mean normal error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        k: int, loss = np.sqrt(loss, 1/k).
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    if log:
        loss = (np.power(np.abs(np.log1p(y_true)-np.log1p(y_pred))*sample_weight, k)).mean()
    else:
        loss = (np.power(np.abs(y_true-y_pred)*sample_weight, k)).mean()
    if root:
        loss = np.power(loss, 1/k)
    return loss


def mean_absolute_error(y_true, y_pred, log=False, sample_weight=None):
    """Mean absolute error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return normal_loss(y_true, y_pred, k=1, log=log, root=False, sample_weight=sample_weight)


def mean_squared_error(y_true, y_pred, log=False, root=False, sample_weight=None):
    """Mean squared error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    return normal_loss(y_true, y_pred, k=2, log=log, root=root, sample_weight=sample_weight)


def mean_absolute_percentage_error(y_true, y_pred, sample_weight=None):
    """Mean absolute percentage error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    value = np.abs(y_pred - y_true)
    y_true[np.where(y_true==0)] = 1
    return (value/np.abs(y_true)*sample_weight).mean()


def hinge(y_true, y_pred, k=1, sample_weight=None):
    """hinge regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        k: int, pow() function dim.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return np.power((1-y_true*y_pred).clip(min=0)*sample_weight, k).mean()


def explained_variance_score(y_true, y_pred, sample_weight=None):
    """explained variance regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return 1-((y_true-y_pred)*sample_weight).std()**2/y_true.std()**2


def median_absolute_error(y_true, y_pred, sample_weight=None):
    """Median absolute error regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return np.median(np.abs(y_true-y_pred)*sample_weight)


def r2_score(y_true, y_pred, sample_weight=None):
    """r2 regression loss.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return 1-np.power((y_true-y_pred)*sample_weight, 2).sum()/np.power((y_true-y_true.mean())*sample_weight, 2).sum()


def report_regression(y_true, y_pred, sample_weight=None, printable=False):
    """regression report
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels.
        sample_weight: list or array of sample weight.
        printable: bool, print report.
    Returns:
        dict, regression report.
    """
    result = {'mean_absolute_error':mean_absolute_error(y_true, y_pred, sample_weight=sample_weight),
              'mean_squared_error':mean_squared_error(y_true, y_pred, sample_weight=sample_weight),
              'mean_absolute_percentage_error':mean_absolute_percentage_error(y_true, y_pred, sample_weight=sample_weight),
              'hinge_loss':hinge(y_true, y_pred, sample_weight=sample_weight),
              'explained_variance_score':explained_variance_score(y_true, y_pred, sample_weight=sample_weight),
              'median_absolute_error':median_absolute_error(y_true, y_pred, sample_weight=sample_weight),
              'r2_score':r2_score(y_true, y_pred, sample_weight=sample_weight)
             }
    if printable:
        print("\nRegression Report")
        print("mean_absolute_error: %.4f" % result['mean_absolute_error'])
        print("mean_squared_error: %.4f" % result['mean_squared_error'])
        print("mean_absolute_percentage_error: %.4f" % result['mean_absolute_percentage_error'])
        print("hinge_loss: %.4f" % result['hinge_loss'])
        print("explained_variance_score: %.4f" % result['explained_variance_score'])
        print("median_absolute_error: %.4f" % result['median_absolute_error'])
        print("r2_score: %.4f" % result['r2_score'])
    return result


def mean_relative_error(y_true, y_pred, normalizer, sample_weight=None):
    """Computes the mean relative error by normalizing with the given values.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        normalizer: The normalizer values with same shape as y_pred.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return (np.abs(y_true-y_pred)/np.array(normalizer)*sample_weight).mean()


def poisson(y_true, y_pred, sample_weight=None):
    """Computes the Poisson loss between y_true and y_pred.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return np.mean((y_pred - y_true * np.log(y_pred+ 1e-7))*sample_weight)


def log_cosh_error(y_true, y_pred, sample_weight=None):
    """Computes the logarithm of the hyperbolic cosine of the prediction error.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a regression.
        sample_weight: list or array of sample weight.
    Returns:
        regression loss values.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    x = (y_pred-y_true)*sample_weight
    sample_weight = np.ones(len(y_true)) if sample_weight is None else np.array(sample_weight)
    sample_weight = sample_weight/sample_weight.sum()*len(sample_weight)
    return np.mean(x + np.log(np.exp(-2. * x) + 1.) - np.log(2.))



# from linora.data._array_dataset import *
# from linora.data._text_line_dataset import *

from linora.data import Dataset
from linora.data import datasets
from linora.data import ImageDataset
from linora.data import TextLineDataset
from linora.data._compress import *
from linora.data._utils import *


import os
import pathlib
from itertools import chain

import numpy as np
import pandas as pd

from linora.data._dataset import DataSet
from linora.parallel._thread import ThreadLoom

__all__ = ['Dataset', 'ImageDataset']


class Dataset(DataSet):
    """Represents a potentially large set of elements from dataframe."""
    def __init__(self):
        super(Dataset, self).__init__()
        
    def from_tensor(self, data):
        """Creates a Dataset whose elements are slices of the given tensors.
        
        Args:
            data: numpy array or tuple of numpy array
        """
        if isinstance(data, tuple):
            for i in data:
                assert len(data[0])==len(i), 'Length needs to be consistent between data.'
        self._params_init()
        self.params.data_mode = 'list' if isinstance(data, tuple) else 'array'
        self.params.data_index = list(range(len(data[0] if isinstance(data, tuple) else data)))
        self.params.data = [np.array(i) for i in data] if isinstance(data, tuple) else np.array(data)
        return self
    
    def _batch_list_map(self, loc):
        data = list(map(self.params.map_func, *(i[loc] for i in self.params.data)))
        return [np.array(list(map(lambda x:x[i], data))) for i in range(len(data[0]))]
    
    def _batch_list(self, loc):
        return [i[loc] for i in self.params.data]
    
    def _batch_map(self, loc):
        return np.array(list(map(self.params.map_func, self.params.data[loc])))
    
    def _batch(self, loc):
        return self.params.data[loc]


class ImageDataset(DataSet):
    """Represents a potentially large set of elements from image path."""
    def __init__(self):
        super(ImageDataset, self).__init__()
        
    def from_tensor(self, data):
        """Creates a Dataset whose elements are slices of the given tensors.
        
        Args:
            data: image path or tuple of image path
        """
        if isinstance(data, tuple):
            for i in data:
                assert len(data[0])==len(i), 'Length needs to be consistent between data.'
        self._params_init()
        self.params.data_mode = 'list' if isinstance(data, tuple) else 'image'
        self.params.data_index = list(range(len(data[0] if isinstance(data, tuple) else data)))
        self.params.data = [np.array(i) for i in data] if isinstance(data, tuple) else np.array(data)
        return self
    
    def from_folder(self, root, image_format=['png', 'jpg', 'jpeg'], label_func=None, label_encoder=False):
        """Construct an image dataset label index.

        Args:
            root: image dataset file root.
            image_format: list, default ['png', 'jpg', 'jpeg'].
            label_func: Function is applied to the name of each picture.
            label_encoder: whether encode labels with value between 0 and n_classes-1.
        """
        self._params_init()
        p = pathlib.Path(root)
        dataset = pd.DataFrame(chain.from_iterable((p.rglob(f'*.{i}') for i in image_format)), columns=['image'])
        if label_func is not None:
            dataset['label'] = dataset.image.map(lambda x:label_func(x.name))
            name_label_dict = {j: i for i, j in enumerate(dataset.label.unique())}
            self.name_label_dict = {'positive':name_label_dict, 'negative':{j: i for i, j in name_label_dict.items()}}
            if label_encoder:
                dataset['label'] = dataset.label.replace(self.name_label_dict['positive'])
        dataset['image'] = dataset.image.astype(str).map(lambda x:eval(repr(x).replace("\\", '/').replace("//", '/')))
        if label_func is not None:
            self.params.data = [dataset.image.values, dataset.label.values]
        else:
            self.params.data = dataset.image.values
        self.params.data_index = dataset.index.to_list()
        
    def from_class_folder(self, root, image_format=['png', 'jpg', 'jpeg'], label_encoder=False):
        """Construct an image dataset label index.
        
        Args:
            root: image dataset file root.
            image_format: list, default ['png', 'jpg', 'jpeg'].
            label_encoder: whether encode labels with value between 0 and n_classes-1.
        """
        self._params_init()
        file = os.listdir(root)
        file = [i for i in file if os.path.isdir(root+'/'+i) and i[0]!='.']
        data = pd.DataFrame()
        for i in file:
            data = pd.concat([data, pd.DataFrame({'image':os.listdir(root+'/'+i), 'label':i})])
        data = data.reset_index(drop=True)
        data['image'] = root+'/'+data.label+'/'+data.image
        data = data[data.image.map(lambda x: True if '.' in x.split('/')[-1] else False)]
        data = data[data.image.map(lambda x: True if x.split('/')[-1][0]!='.' else False)]
        data = data[data.image.map(lambda x: True if len(x.split('/')[-1].split('.'))==2 else False)]
        data = data[data.image.map(lambda x: True if str.lower(x.split('/')[-1].split('.')[1]) in image_format else False)]
        data = data.reset_index(drop=True)
        name_label_dict = {j: i for i, j in enumerate(data.label.unique())}
        self.name_label_dict = {'positive':name_label_dict, 'negative':{j: i for i, j in name_label_dict.items()}}
        if label_encoder:
            data['label'] = data.label.replace(self.name_label_dict['positive'])
        self.params.data = [data.image.values, data.label.values]
        self.params.data_index = data.index.to_list()

    def _batch_list_map(self, loc):
        loom = ThreadLoom(self.params.map_size)
        for i in loc:
            loom.add_function(self.params.map_func, [j[i] for j in self.params.data])
        t = loom.execute()
        for i in t:
            if t[i]['got_error']:
                continue
            if isinstance(t[i]['output'], (list, tuple)):
                return [np.concatenate([np.expand_dims(t[j]['output'][k], 0) for j in t if not t[j]['got_error']]) for k in range(len(t[i]['output']))]
            else:
                return np.concatenate([np.expand_dims(t[j]['output'], 0) for j in t if not t[j]['got_error']])
    
    def _batch_list(self, loc):
        return [i[loc] for i in self.params.data]
    
    def _batch_map(self, loc):
        loom = ThreadLoom(self.params.map_size)
        for i in loc:
            loom.add_function(self.params.map_func, [self.params.data[i]])
        t = loom.execute()
        print(i)
        for i in t:
            if t[i]['got_error']:
                continue
            if isinstance(t[i]['output'], (list, tuple)):
                return [np.concatenate([np.expand_dims(t[j]['output'][k], 0) for j in t if not t[j]['got_error']]) for k in range(len(t[i]['output']))]
            else:
                return np.concatenate([np.expand_dims(t[j]['output'], 0) for j in t if not t[j]['got_error']])
    
    def _batch(self, loc):
        return self.params.data[loc]